#!/usr/bin/python
# -*- coding: utf-8 -*-
import hashlib
import hmac
import logging
import urllib
import urlparse

import M2Crypto
import datetime
import requests

logger = logging.getLogger(__name__)


def get_access_key_from_pem(cert):
    cert = M2Crypto.X509.load_cert(cert)
    return str(cert.get_fingerprint('sha256')).zfill(64).lower()


def get_secret_key_from_pem(cert):
    cert = M2Crypto.X509.load_cert(cert)
    return str(cert.get_fingerprint('sha512')).zfill(128).lower()[:64]


class ZuulClient(object):
    def __init__(self, endpoint, zuul_pem, client_pem):
        self.access_key = get_access_key_from_pem(client_pem[0])
        self.secret_key = get_secret_key_from_pem(client_pem[0])
        self.zuul = ZuulConnector(endpoint, self.access_key, self.secret_key,
                                  verify=zuul_pem, cert=client_pem)

    def raise_error_on_failure(self, response):
        if response.ok:
            return
        try:
            response.raise_for_status()
        except Exception as e:
            logger.exception(e)
            logger.error(response.content)
            print response.content
            raise e

    def get_api_keys(self, user, region):
        uri = 'credentials'
        args = {'region': region, 'primaryAccessKey': user}
        response = self.zuul.call(args=args, uri=uri, method='get')
        self.raise_error_on_failure(response)
        return response.json()

    def create_api_keys(self, region, primary_access_key, access_key,
                        secret_key):
        uri = 'credentials'
        args = {'region': region, 'primaryAccessKey': primary_access_key,
                'accessKey': access_key, 'secretKey': secret_key}
        response = self.zuul.call(args=args, uri=uri, method='post')
        self.raise_error_on_failure(response)
        return response.json()

    def update_api_keys(self, region, primary_access_key, access_key,
                        new_secret_key):
        uri = 'credentials'
        args = {'region': region, 'primaryAccessKey': primary_access_key,
                'accessKey': access_key, 'newSecretKey': new_secret_key}
        response = self.zuul.call(args=args, uri=uri, method='put')
        self.raise_error_on_failure(response)
        return response.json()

    def delete_api_keys(self, region, primary_access_key):
        uri = 'credentials'
        args = {'region': region, 'primaryAccessKey': primary_access_key}
        response = self.zuul.call(args=args, uri=uri, method='delete')
        self.raise_error_on_failure(response)
        return response.json()

    def get_regions(self):
        uri = 'regions'
        response = self.zuul.call(uri=uri, method='get')
        self.raise_error_on_failure(response)
        return response.json()

    def create_region(self, name, description):
        uri = 'regions'
        args = {'name': name, 'description': description}
        response = self.zuul.call(uri=uri, args=args, method='post')
        return response.json()


class ZuulConnector(object):
    def __init__(self, url, api_key, secret_key, region='hybrid', cert=None,
                 verify=False):
        """api_key and secret keys are used for
         the authentication part of the request"""
        self.api_key = api_key
        self.secret_key = str(
            secret_key)  # str cast-> hmac cant handle unicode
        self.url = self.append_slash(url)
        self.service = 'ec2'
        self.region = region
        self.cert = cert
        self.verify = verify

    def append_slash(self, url):
        if url.endswith('/'):
            return url
        return url + '/'

    def call(self,
             args={'PrimaryApiKey': None, 'AccessKey': None, 'SecretKey': None,
                   'Region': None, 'Version': '2014-10-01'}, method='POST',
             uri='/'):
        query = self.build_query_string(args)
        t = datetime.datetime.utcnow()
        date_stamp = self.get_date_stamp(t)
        amzdate_stamp = self.get_amzdate_stamp(t)
        request_parameters, headers = self.make_request(
            method, uri, query, date_stamp, amzdate_stamp)
        return self.send_request(method, self.url + uri + '?' + query,
                                 request_parameters, headers)

    def send_request(self, method, url, request_parameters, headers):
        http_method = getattr(requests, method.lower())
        return http_method(url, data=request_parameters, headers=headers,
                           cert=self.cert, verify=self.verify)

    def build_query_string(self, args):
        params = []
        for k in sorted(args.keys()):
            if args[k]:
                params.append(k + '=' + urllib.quote_plus(
                    unicode(args[k]).encode('utf-8')).replace("+", "%20"))
        query = '&'.join(params)
        return query

    def get_amzdate_stamp(self, dt):
        return dt.strftime('%Y%m%dT%H%M%SZ')

    def get_date_stamp(self, dt):
        return dt.strftime('%Y%m%d')

    def build_request(self, method, uri, query, headers, signed_headers,
                      payload_hash):
        return method.upper() + '\n/{}\n{}\n' \
                                '{}\n{}\n{}'.format(uri, query, headers,
                                                    signed_headers,
                                                    payload_hash)

    def make_request(self, method, uri, query, datestamp, amzdate):
        url = urlparse.urlparse(self.url)
        headers = 'host:' + url.netloc + '\n' + 'x-amz-date:' + amzdate + '\n'
        signed_headers = 'host;x-amz-date'
        payload_hash = hashlib.sha256(query).hexdigest()
        request = self.build_request(method, uri, query, headers,
                                     signed_headers, payload_hash)
        algorithm = 'AWS4-HMAC-SHA256'
        credential_scope = '{0}/{1}/{2}/aws4_request'.format(
            datestamp,
            self.region,
            self.service
        )

        string_to_sign = '{}\n{}\n{}\n{}'.format(
            algorithm, amzdate,
            credential_scope, hashlib.sha256(request).hexdigest()
        )
        signing_key = self.get_signature_key(self.secret_key, datestamp,
                                             self.region, self.service)
        signature = hmac.new(signing_key, (string_to_sign).encode('utf-8'),
                             hashlib.sha256).hexdigest()

        authorization_header = "{} Credential={}/{}, SignedHeaders={}, Signature={}".format(  # noqa
            algorithm, self.api_key, credential_scope, signed_headers,
            signature)
        headers = {'x-amz-date': amzdate,
                   'Authorization': authorization_header,
                   'content-type': 'application/x-www-form-urlencoded'}
        return (query, headers)

    def sign(self, key, msg):
        return hmac.new(key, msg.encode('utf-8'), hashlib.sha256).digest()

    def get_signature_key(self, key, date_stamp, region_name, service_name):
        date_key = self.sign(('AWS4' + key).encode('utf-8'), date_stamp)
        region_key = self.sign(date_key, region_name)
        service_key = self.sign(region_key, service_name)
        return self.sign(service_key, 'aws4_request')
